// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import C25519Cipher from './c25519Cipher';
import AesSherableKey from './aesSherableKey';

export default class CompositeCipher {
    /**
     * @param {C25519Cipher} key
     * @param {Uint8Array} cipher
     */
    constructor(key, cipher) {
        this.cipherKey = key;
        this.cipher = cipher;
    }

    /**@param {import("./c25519Key").default} prvKey */
    decrypt(prvKey) {
        const key = AesSherableKey.from(this.cipherKey.decrypt(prvKey));
        return key.decrypt(this.cipher);
    }

    toString() {
        return this.toArray().toString('base64');
    }
    
    toArray() {
        const bufferKey = this.cipherKey.toArray();

        const buffer = Buffer.alloc(bufferKey.length + this.cipher.length);
        buffer.set(bufferKey, 0);
        buffer.set(this.cipher, bufferKey.length);
        
        return buffer;
    }
    
    /**@param {Uint8Array | string} data
     * @param {import("./c25519Key").default} key */
    static encrypt(data, key) {
        const randomKey = new AesSherableKey();
        const cipherData = randomKey.encrypt(data);
        const cipherKey = key.encryptKey(randomKey);

        return new CompositeCipher(cipherKey, cipherData);
    }

    /** @param {Uint8Array|string} data */
    static from(data) {
        if (!data || data.length < 0)
            throw Error("Empty cipher is not valid");
        
        const cipher = fromBase64(data);
        if (cipher.length <= 96)
            throw Error("Cipher is not in correct format");
        
        return new CompositeCipher(C25519Cipher.from(cipher.slice(0, 96)), cipher.slice(96));
    }
}

/** 
 * @param {string|Uint8Array} data
 * @return {Uint8Array}
 */
function fromBase64(data) {
    return typeof data === "string" ? Buffer.from(data, "base64") : data;
}