
import BI from 'big-integer';
import C25519Point from './c25519Point';
import Hash from './hash';



const _0n = BigInt(0);
const _1n = BigInt(1);
const _2n = BigInt(2);
const _5n = BigInt(5);
const _8n = BigInt(8);

export default class ed25519Point {

    static get a() {return BigInt(-1)};
    static get d() {return BigInt("37095705934669439343138083508754565189542113879843219016388785533085940283555");}
    static get d_FOR_Y_FROM_X() {return BigInt("20800338683988658368647408995589388737092878452977063003340006470870624536394");}
    static get p() {return BigInt("57896044618658097711785492504343953926634992332820282019728792003956564819949");}
    static get order() { return BigInt("7237005577332262213973186563042994240857116359379907606001950938285454250989");}
    static get g() { return new ed25519Point(
        BigInt("15112221349535400772501151409588531511454012693041857206046113283949847762202"),
        BigInt("46316835694926478169428394003475163141307993866256225615783033603165251855960"), 
        _1n, 
        BigInt("46827403850823179245072216630277197565144205554125654976674165829533817101731")
    );}
    static get infinity() {return new ed25519Point(_0n, _1n, _1n, _0n);} //infinity also known as identity for ed25519
    static get SQRT_M1() {return BigInt("19681161376707505956807079304988542015446066515923890162744021073123829784752");}
    static get WIER_POINT_CONV() {return BigInt("18783869910803591432408634235432691537229359092449989392190168916153006498937509046");}
    static get sqrt_minus_486664() {return BigInt("51042569399160536130206135233146329284152202253034631822681833788666877215207");} // one of the square roots of -486664 in finite field p

    constructor(x, y=null, z=null, t=null) {
        this.x = x;
        this.y = y !== null ? y : getYfromX(y);
        this.z = z !== null ? z : _1n;
        this.t = t !== null ? t : mod(this.x * this.y);
    }
    

    isInfinity(){
        return this.isEqual(ed25519Point.infinity);  // TODO: Check if  t needs check too.
    }

    isEqual(other){
        const X1Z2 = mod(this.x * other.z);
        const X2Z1 = mod(other.x * this.z);
        const Y1Z2 = mod(this.y * other.z);
        const Y2Z1 = mod(other.y * this.z);
        return X1Z2 === X2Z1 && Y1Z2 === Y2Z1;
    }

    negate(){
        return new ed25519Point(mod(-this.x), this.y, this.z, mod(-this.t)); //// not checked if works
    }


    // Using double and add algorithm.
    times(num) {
        num = BigInt(num.toString());
        var point = new ed25519Point(this.x, this.y, this.z, this.t);
        let newPoint = new ed25519Point(_0n, _1n, _1n, _0n); // identity aka infinity
        while (num > _0n) {
            if ((num & _1n) === (_1n)) {
                newPoint = newPoint.add(point);
            }
            point = point.double();
            num = num >> _1n;
        }
        return newPoint;
    }

    double() {
        let A = mod(this.x * this.x);
        let B = mod(this.y * this.y);
        let C = mod(_2n * mod(this.z * this.z));
        let D = mod(ed25519Point.a * A);
        let x1y1 = this.x + this.y;
        let E = mod(mod(x1y1 * x1y1) - A - B);
        let G = D + B;
        let F = G - C;
        let H = D - B;
        let X3 = mod(E * F);
        let Y3 = mod(G * H);
        let T3 = mod(E * H);
        let Z3 = mod(F * G);
        return new ed25519Point(X3, Y3, Z3, T3);
    }

    add(other) {
        let A = mod((this.y - this.x) * (other.y + other.x));
        let B = mod((this.y + this.x) * (other.y - other.x));
        let F = mod(B - A);
        if (F == _0n) return this.double();
        let C = mod(this.z * _2n * other.t);
        let D = mod(this.t * _2n * other.z);
        let E = D + C;
        let G = B + A;
        let H = D - C;
        let X3 = mod(E * F);
        let Y3 = mod(G * H);
        let T3 = mod(E * H);
        let Z3 = mod(F * G);
        return new ed25519Point(X3, Y3, Z3, T3);
    }

    // Notes on getX and getY for next dev who runs by this.
    // From what I found, using the native BigInt lib is MUCH easier (and faster apparently)
    // to perform all the maths above compared to big-integer lib.
    // Unfortunately, most of the signing operations in cryptide 
    // use big-integer as it's easier to convert to bytes etc.
    // So its important to stay consistent and use big-integer everhwere.
    // This is why the getX returns a big-integer object, converted
    // from a BigInt object.
    getX(){
        return BI(mod(this.x * mod_inv(this.z)).toString());
    }
    getY(){
        return BI(mod(this.y * mod_inv(this.z)).toString());
    }
    /**
     * @param {Uint8Array} data
     */
    static from(data){ 
        var x = BigInt(BI.fromArray(Array.from(data.slice(0, 32)), 256, false).toString()); 
        var y = BigInt(BI.fromArray(Array.from(data.slice(32, 64)), 256, false).toString());  
        return new ed25519Point(x, y);
    }

    /** @returns {Uint8Array} */
    toArray(){
        var r_xBuff = this.getX().toArray(256).value;
        while(r_xBuff.length < 32){r_xBuff.unshift(0);} // pad if array < 32
    
        var r_yBuff = this.getY().toArray(256).value;
        while(r_yBuff.length < 32){r_yBuff.unshift(0);} // pad if array < 32  Big endian

        return new Uint8Array(r_xBuff.concat(r_yBuff));
    }

    static fromString(message){
        // this function genereates a pseudo random point (not multiple of G) from a message
        var x = mod(BigInt(BI.fromArray(Array.from(Hash.shaBuffer(message)), 256, false).toString()));

        var point = new ed25519Point(x);

        while (point.times(ed25519Point.order).isInfinity()) {
            x = x + _1n; 
            point = new ed25519Point(x);
        }

        return point;
    }

    static fromX25519(x){ // works
        var ed_y = mod((x - _1n) * mod_inv(x + _1n));
        var ed_x = getXfromY(ed_y);
        return new ed25519Point(ed_x, ed_y);
    }

    static fromX25519_Weierstrass(x){
        var a = mod(x - ed25519Point.WIER_POINT_CONV)
        var ed_y = mod((a - _1n) * mod_inv(a + _1n));
        var ed_x = getXfromY(ed_y);
        return new ed25519Point(ed_x, ed_y);
    }

    toCurve25519_Wierstrass(){
        var c25519_x_mont = mod(mod(this.z + this.y) * mod_inv(this.z - this.y)) // taken from https://crypto.stackexchange.com/questions/9536/converting-ed25519-public-key-to-a-curve25519-public-key
        var c25519_y = mod(ed25519Point.sqrt_minus_486664 * c25519_x_mont * mod_inv(BigInt(this.getX().toString()))); // generate y. taken from link above. keep in mind y in wierstrass and montomgery are the same
        var c25519_x_wierstrass = mod(c25519_x_mont + ed25519Point.WIER_POINT_CONV); // convert montogmery to wierstrass
        return new C25519Point(BI(c25519_x_wierstrass.toString()), BI(c25519_y.toString()));
    }
}

//works
function pow2(x, power){
    let res = x;
    while (power-- > _0n) {
        res = mod(res * res);
    }
    return res;
}
// Power to (p-5)/8 aka x^(2^252-3)
// Used to calculate y - the square root of y².
// Exponentiates it to very big number.
// We are unwrapping the loop because it's 2x faster.
// (2n**252n-3n).toString(2) would produce bits [250x 1, 0, 1]
// We are multiplying it bit-by-bit
/**
 * 
 * @param {bigint} x 
 * @returns 
 */
//works
function pow_2_252_3(x) {
    const P = ed25519Point.p;
    const _5n = BigInt(5);
    const _10n = BigInt(10);
    const _20n = BigInt(20);
    const _40n = BigInt(40);
    const _80n = BigInt(80);
    const x2 = (x * x) % P;
    const b2 = (x2 * x) % P; // x^3, 11
    const b4 = (pow2(b2, _2n) * b2) % P; // x^15, 1111
    const b5 = (pow2(b4, _1n) * x) % P; // x^31
    const b10 = (pow2(b5, _5n) * b5) % P;
    const b20 = (pow2(b10, _10n) * b10) % P;
    const b40 = (pow2(b20, _20n) * b20) % P;
    const b80 = (pow2(b40, _40n) * b40) % P;
    const b160 = (pow2(b80, _80n) * b80) % P;
    const b240 = (pow2(b160, _80n) * b80) % P;
    const b250 = (pow2(b240, _10n) * b10) % P;
    const pow_p_5_8 = (pow2(b250, _2n) * x) % P;
    // ^ To pow to (p+3)/8, multiply it by x.
    return {pow_p_5_8, b2};
}
// this method is incomplete. not ready for ed25519 point decoding
function uvRatio(u, v) {
    const v3 = mod(v * v * v);                  // v³
    const v7 = mod(v3 * v3 * v);                // v⁷
    const pow = pow_2_252_3(u * v7).pow_p_5_8;
    let x = mod(u * v3 * pow);                  // (uv³)(uv⁷)^(p-5)/8
    const vx2 = mod(v * x * x);                 // vx²
    const root1 = x;                            // First root candidate
    const root2 = mod(x * ed25519Point.SQRT_M1);             // Second root candidate
    const useRoot1 = vx2 === u;                 // If vx² = u (mod p), x is a square root
    const useRoot2 = vx2 === mod(-u);           // If vx² = -u, set x <-- x * 2^((p-1)/4)
    const noRoot = vx2 === mod((_0n-u) * ed25519Point.SQRT_M1);   // There is no valid root, vx² = -u√(-1)
    if (useRoot1) x = root1;
    if (useRoot2 || noRoot) x = root2;          // We return root2 anyway, for const-time
    if (edIsNegative(x)) x = mod(-x);
    return { isValid: useRoot1 || useRoot2, value: x };
  }
// Little-endian check for first LE bit (last BE bit);
function edIsNegative(num) {
    return (mod(num) & _1n) === _1n;
}
/**
 * 
 * @param {bigint} x 
 * @returns 
 */
function getYfromX(x){
    // Always get 'higher' point. Remember there are two possible values, this always picks the first one to remain consistent.
    var x2 = mod(x * x);
    var u = mod(_1n + x2);
    var v = mod(_1n + ed25519Point.d_FOR_Y_FROM_X*x2);

    return uvRatio(u, v).value;
}

// Will always be 'higher' x to ensure keys are the same when transitioning fro curve25519 to ed25519
function getXfromY(y){
    const y2 = mod(y * y);
    const u = mod(y2 - _1n);
    const v = mod(ed25519Point.d * y2 + _1n);
    return uvRatio(u, v).value;
}

// ensure the X coordinate is supplied here




function mod(a, b = ed25519Point.p) {
    var res = a % b;
    return res >= _0n ? res : b + res;
}

function mod_inv(number, modulo = ed25519Point.p) {
    if (number === _0n || modulo <= _0n) {
        throw new Error(`invert: expected positive integers, got n=${number} mod=${modulo}`);
    }
    let a = mod(number, modulo);
    let b = modulo;
    // prettier-ignore
    let x = _0n, y = _1n, u = _1n, v = _0n;
    while (a !== _0n) {
        const q = b / a;
        const r = b % a;
        const m = x - u * q;
        const n = y - v * q;
        // prettier-ignore
        b = a, a = r, x = u, y = v, u = m, v = n;
    }
    const gcd = b;
    if (gcd !== _1n) throw new Error('invert: does not exist');
    return mod(x, modulo);
}




