// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import BigInt from 'big-integer';
import utils from '../utils';
import PaillierKey from './paillierKey';
import PaillierScalar from './paillierScalar';
import PHelp from '../phelp';

export default class PaillierCipher {

    /**
     * @param {bigInt.BigInteger} m
     * @param {PaillierKey} key
     */
    static encrypt(m, key) {
        return PHelp.run(function enc() {
            var r = utils.random(BigInt.one, key.n.subtract(BigInt.one));
            var c = r.modPow(key.n, key.nsquare).times(key.n.times(m).add(BigInt.one)).mod(key.nsquare)

            return new PaillierScalar(c, key);
        });
    }

    /**
     * @param {PaillierScalar} scalar
     * @param {PaillierKey} key
     */
    static decrypt(scalar, key) {
        if (!key.isPrivate)
            throw new Error("Private key is not valid");

        return PHelp.run(function dec() {
            var mp = PaillierCipher.evalL(scalar.c, key.p, key.psquare).times(key.hp).mod(key.p);
            var mq = PaillierCipher.evalL(scalar.c, key.q, key.qsquare).times(key.hq).mod(key.q);
            
            return chinese([mp, mq], [key.p, key.q]).mod(key.n);
        });
    }

    /**
     * @param {bigInt.BigInteger} g 
     * @param {bigInt.BigInteger} p 
     * @param {bigInt.BigInteger} psquare 
     */
    static evalL(g, p, psquare) {
        return g.modPow(p.minus(BigInt.one), psquare).minus(BigInt.one).divide(p);
    }
}

/**
 * 
 * @param {bigInt.BigInteger[]} bs 
 * @param {bigInt.BigInteger[]} ns 
 */
function chinese(bs, ns)
{
    var Ns = [ ns[1], ns[0] ];
    var xs = [ utils.primeInv(Ns[0], ns[0]), utils.primeInv(Ns[1], ns[1]) ];
    var rs = [ bs[0].times(Ns[0]).times(xs[0]), bs[1].times(Ns[1]).times(xs[1]) ];
    var x = rs[0].add(rs[1]).mod(ns[0].times(ns[1]));
    return x;
}
